from lib2to3.pgen2.token import LEFTSHIFT
from bacon.net import baconNet, expression, term
import tensorflow as tf

import sys
sys.path.append("..")


class poly2(baconNet):
    def __init__(self, optimizer='adam', initializer='identity'):
        super().__init__(6, optimizer=optimizer, initializer=initializer)

    def explain_contribution(self, m, c, singleVariable=False):
        a = 1
        b = 1
        if singleVariable:
            b = 0
        delta = 0.01
        terms = []
        if abs(m[2][0]) > delta and abs(a) > delta:
            terms.append(term(coefficient=m[2][0], term="[x]^2"))
        if abs(m[4][0]) > delta and abs(b*a) > delta:
            terms.append(term(coefficient=m[4][0], term="[x][y]"))
        if abs(m[3][0]) > delta and abs(b) > delta:
            terms.append(
                term(coefficient=m[3][0], term="[y]^2"))
        if abs(m[0][0]) > delta and abs(a) > delta:
            terms.append(term(coefficient=m[0][0], term="[x]"))
        if abs(m[1][0]) > delta and abs(b) > delta:
            terms.append(term(coefficient=m[1][0], term="[y]"))
        if abs(m[5][0]) > delta and abs(c) > delta:
            terms.append(term(coefficient=c))
        return expression(terms)

    def expand(self, a, b):
        X = tf.stack((tf.cast(a, dtype='float32'),
                      tf.cast(b, dtype='float32'),
                      tf.cast(tf.math.pow(a, 2), dtype='float32'),
                      tf.cast(tf.math.pow(b, 2), dtype='float32'),
                      tf.cast(tf.math.multiply(a, b), dtype='float32'),
                      tf.ones(tf.shape(a), dtype='float32')))
        return tf.transpose(X)



from abc import abstractmethod


from abc import ABC
import tensorflow as tf
from tensorflow import keras
from keras.utils.generic_utils import get_custom_objects
from keras.layers import Activation
import numpy as np
from sympy import *
# surpress warnings on running GPU Tensorflow on CPU
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'


class baconNet(ABC):
    def __custom_activation(self, x):
        return x

    def __init__(self, size, constantTerm=True, optimizer='adam', initializer='identity'):
        get_custom_objects().update(
            {'custom_activation': Activation(self.__custom_activation)})
        self.__model = tf.keras.models.Sequential()
        self.__model.add(tf.keras.Input(shape=(2,)))
        self.__model.add(expansionLayer(self.expand))
        self.__model.add(tf.keras.layers.Dense(
            size, activation=self.__custom_activation, kernel_initializer=initializer, bias_initializer='zeros'))
        self.__model.add(tf.keras.layers.Dense(
            1, activation=self.__custom_activation))
        self.__model.compile(
            optimizer=optimizer, loss='mean_squared_error', metrics=['mae'])
        self.__constantTerm = constantTerm
        self.__size = size

    @ abstractmethod
    def explain_contribution(self, m, c, singleVariable=False):
        pass

    def explain(self, singleVariable=False):
        return self.explain_model(self.__model, singleVariable)

    def explain_model(self, model, singleVariable=False):
        m, c = self.get_contribution_from_model(
            self.__model, self.__size, self.__constantTerm)
        return self.explain_contribution(m, c, singleVariable)

    def get_contribution_from_model(self, model, size, constantTerm=True):
        m = np.matmul(model.layers[1].weights[0], model.layers[2].weights[0])
        if constantTerm:
            # assuming the last input is a constant column
            bias0 = model.layers[1].bias.numpy()
            weight0 = model.layers[1].weights[0][size-1]
            for i in range(len(bias0)):
                bias0[i] = bias0[i] + weight0[i]
            weight1 = model.layers[2].weights[0]
            for i in range(len(bias0)):
                bias0[i] = bias0[i] * weight1[i][0]
            c = bias0.sum() + model.layers[2].bias.numpy()[0]
        else:
            c = 0
        return m, c

    @ abstractmethod
    def expand(self, a, b):
        pass

    def predict(self, a, b=0):
        simple = False
        if isinstance(a, list):
            if isinstance(b, list):
                x = np.column_stack((a, b))
            else:
                x = np.column_stack((a, np.zeros(len(a))))
        else:
            simple = True
            x = np.column_stack(([a], [b]))
        if simple:
            return self.__model.predict(x)[0][0]
        else:
            return self.__model.predict(x).flatten()

    def fit(self, a, b, y):
        callback = tf.keras.callbacks.EarlyStopping(monitor='mae', patience=20)
        history = self.__model.fit(
            np.column_stack((a, b)), y, epochs=600, batch_size=10, verbose=0, callbacks=[callback])
        return history

    def get_model(self):
        return self.__model


class dataCreator:
    def create(size, scale, aggregate, params=2):
        values = []
        for i in range(params):
            values.append(np.random.rand(size) * scale)
        if len(values) < 2:
            values.append(np.zeros(size))
        y = aggregate(values)
        return values, y


class expansionLayer(tf.keras.layers.Layer):
    def __init__(self, expander, **kwargs):
        super().__init__(**kwargs)
        self.expander = expander

    def call(self, inputs):
        X = self.expander(
            inputs[:, 0:1], inputs[:, 1:2])
        return X[0]


class expression:
    def __init__(self, terms):
        self.terms = terms

    def string(self, precision=-1, ignoreOne=False, delta=0.01, namePattern='x', patternIndex=0, sympy=True):
        strs = []
        for t in self.terms:
            strs.append(t.string(precision, ignoreOne,
                        delta, namePattern, patternIndex, sympy))
        ret = " + ".join(filter(None, strs))
        return ret

    def simplify(self, exp, precision=0, ignoreOne=True):
        ex1 = simplify(exp.replace("+ -", "-"))
        ex1 = expand(ex1)
        ex2 = ex1
        if precision >= 0:
            for a in preorder_traversal(ex1):
                if isinstance(a, Float):
                    ex2 = ex2.subs(a, round(a, precision))
        strExp = str(simplify(ex2))
        strExp = strExp.replace('**', '^')
        if precision >= 0 and ignoreOne:
            strExp = strExp.replace(str(round(1.0, precision))+'*', '')
        strExp = strExp.replace('*', '')
        return strExp


class term:
    def __init__(self, coefficient=0, term="", leftExp=None):
        self.coefficient = coefficient
        self.term = term
        self.leftExp = leftExp

    def string(self, precision=-1, ignoreOne=False, delta=0.01, namePattern='x', patternIndex=0, sympy=True):
        if self.coefficient == 0:
            return ""
        if ignoreOne and abs(self.coefficient-1) < delta and (not sympy):
            ret = ""
        else:
            if precision == -1:
                ret = str(self.coefficient)
            else:
                ret = str(round(self.coefficient, precision))
        txt = self.term
        if sympy:
            txt = txt.replace("][", "]*[")
        if self.leftExp != None:
            txt = txt.replace('[x]', "(" + self.leftExp.string(
                precision, ignoreOne, delta, namePattern, patternIndex-2, sympy) + ")")
        if namePattern != 'x1':
            txt = txt.replace('[x]', chr(ord(namePattern)+patternIndex))
            txt = txt.replace('[y]', chr(ord(namePattern)+patternIndex+1))
        else:
            txt = txt.replace('[x]', 'x{' + str(patternIndex) + "}")
            txt = txt.replace('[y]', 'x{' + str(patternIndex+1) + "}")
        if sympy:
            if txt == "":
                return ret
            else:
                return ret + "*" + txt.replace("^", "**")
        else:
            return ret + txt



from bacon.nets.lsp3 import lsp3
from bacon.nets.poly2 import poly2
from tensorflow import keras
from tensorflow.keras import layers
import numpy as np
import tensorflow as tf


class baconStack():
    def __createBacon(self, type, optimizer='adam', initializer='identity'):
        if type == "poly2":
            return poly2(optimizer, initializer)
        elif type == "lsp3":
            return lsp3(optimizer, initializer)
        else:
            raise Exception("unsupported bacon type: " + type)

    def __init__(self, size=0, bacons=None, baconType=None, baconNames=None, optimizer='adam', initializer='identity'):
        self.__bacons = []
        if baconType != None and size > 0:
            for i in range(size):
                self.__bacons.append(self.__createBacon(
                    baconType, optimizer, initializer))
        if baconNames != None:
            if len(self.__bacons) > 0:
                raise Exception("can't use both baconType and baconNames")
            for b in range(baconNames):
                self.__bacons.append(
                    self.__createBacon(b, optimizer, initializer))
        if bacons != None:
            if len(self.__bacons) > 0:
                raise Exception(
                    "bacons parameter can't be used with baconType or baconNames")
            self.__bacons = bacons.copy()
        inputs = []
        i = 3
        for b in self.__bacons:
            if len(inputs) == 0:
                inputs.append(keras.Input(
                    shape=(2,), name="input 1 and 2"))
                x = b.get_model()(inputs[len(inputs)-1])
            else:
                inputs.append(keras.Input(shape=(1,), name="input " + str(i)))
                x = layers.concatenate([x, inputs[len(inputs)-1]])
                x = b.get_model()(x)
                i += 1
        self.__model = keras.Model(inputs=inputs, outputs=[
                                   x], name="bacon-stack")
        self.__model.compile(
            optimizer=optimizer, loss='mean_squared_error', metrics=['mae'])

    def fit(self, x, y):
        cols = len(x)
        rows = len(x[0])
        inputs = []
        if cols < 2:
            inputs.append(np.column_stack(
                (x[0].reshape(-1, 1).flatten(), np.zeros(rows))))
        else:
            inputs.append(np.column_stack(
                (x[0].reshape(-1, 1).flatten(), x[1].reshape(-1, 1).flatten())))
        for i in range(2, cols):
            inputs.append(x[i].reshape(-1, 1).flatten())
        callback = tf.keras.callbacks.EarlyStopping(monitor='mae', patience=20)
        history = self.__model.fit(
            x=[inputs],
            y=[y],
            epochs=600, batch_size=10, verbose=0, callbacks=[callback]
        )
        return history

    def fit2(self, x, y, maeTarget=0.01, maxTries=-1):
        counter = 1
        while True:
            history = self.fit(x.copy(), y.copy())
            mae = history.history['mae'][len(history.history['mae'])-1]
            print(f"attempt {counter}, mae = {mae}")
            if abs(mae) <= maeTarget or maxTries > 0 and counter >= maxTries:
                return history, mae, counter
            counter += 1

    def explain(self):
        baconIdx = 0
        expressions = []
        for idx in range(len(self.__model.layers)):
            layer = self.__model.get_layer(index=idx)
            if type(layer) is keras.Sequential:
                expressions.append(self.__bacons[baconIdx].explain())
                baconIdx += 1
        for i in range(len(expressions)-1):
            for j in range(len(expressions[i+1].terms)):
                expressions[i+1].terms[j].leftExp = expressions[i]
            pass
        return expressions[len(expressions)-1]

    def get_model(self):
        return self.__model



from bacon.net import baconNet, expression, term
import tensorflow as tf

import sys
sys.path.append("..")


class lsp3(baconNet):
    def __init__(self, optimizer='adam', initializer='identity'):
        super().__init__(3, constantTerm=False, optimizer=optimizer, initializer=initializer)

    def explain_contribution(self, m, c, singleVariable=False):
        if singleVariable:
            return expression(terms=[term(term="[x]", coefficient=1.0)])
        delta = 0.01
        terms = []
        if abs(m[0][0]-1) < delta:
            return expression(terms=[term(term="min([x],[y])", coefficient=1.0)])
        if abs(m[1][0]-1) < delta:
            return expression(terms=[term(term="max([x],[y])", coefficient=1.0)])
        if abs(m[0][0]-0.5) < delta and abs(m[1][0]-0.5) < delta:
            return expression(terms=[term(term="([x]+[y])/2", coefficient=1.0)])
        if abs(m[2][0]-1) < delta:
            return expression(terms=[term(term="[x]*[y]", coefficient=1.0)])
        print(m)
        return expression(terms=[term(term="can't explain", coefficient=1.0)])

    def expand(self, a, b):
        X = tf.stack((tf.cast(tf.minimum(a, b), dtype='float32'),
                      tf.cast(tf.maximum(a, b), dtype='float32'),
                      tf.cast(tf.math.multiply(a, b), dtype='float32')))
        return tf.transpose(X)



from setuptools import setup
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

VERSION = '0.1.9'
DESCRIPTION = 'A neural network architecture for building fully explainable neural network for arithmetic and gradient logic expression approximation.'

# Setting up
setup(
    name="bacon-net",
    version=VERSION,
    url="https://github.com/Haishi2016/bacon-net",
    author="haishibai (Haishi Bai)",
    author_email="<haishi.bai@live.com>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    install_requires=[
        "tensorflow>=2.6.2",
        "keras>=2.6.0",
        "numpy>=1.19.5",
        "sympy"
    ],
    keywords=['ai', 'explainable', 'gradient logic',
              'approximation', 'formula', 'explainability', 'decision', 'decision making'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: Apache Software License",
    ],
    extras_require={
        "dev": [
            "pytest>=3.7",
        ],
    },
    packages=["bacon", "bacon.nets"],
    py_modules=["net", "stack", "nets"]
)
